# -*- coding: utf-8 -*-
"""
    torchrecorder.renderer.__init__
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    :copyright: (c) 2020 by Gautham Venkatasubramanian.
    :license: see LICENSE for more details.
"""
from .gv import GraphvizRenderer, GraphvizStyler
